from __future__ import absolute_import
from .technique import PositionInvariantTechnique
import cv2
import numpy as np

class bilateralBlurringAugmentationTechnique(PositionInvariantTechnique):

    # Examples for values of diameter, sigmaColor, and sigmaSpace are
    # (11,21,7), (11,41,21), (11,61,39).
    def __init__(self,parameters):
        PositionInvariantTechnique.__init__(self, parameters)
        if 'diameter' in list(parameters.keys()):
            self.diameter = int(parameters["diameter"])
        else:
            self.diameter = 11
        if 'sigmaColor' in list(parameters.keys()):
            self.sigmaColor = int(parameters["sigmaColor"])
        else:
            self.sigmaColor = 21
        if 'sigmaSpace' in list(parameters.keys()):
            self.sigmaSpace = int(parameters["sigmaSpace"])
        else:
            self.sigmaSpace = 7



    def apply(self, image):
        blurred = cv2.bilateralFilter(image, self.diameter, self.sigmaColor, self.sigmaSpace)
        return blurred


from __future__ import absolute_import
from .iaugmentor import IAugmentor
from abc import ABCMeta
from future.utils import with_metaclass

class IAugmentorClassification(with_metaclass(ABCMeta, IAugmentor)):
    def __init__(self,inputPath,outputPath):
        self.inputPath = inputPath
        self.outputPath = outputPath




from __future__ import absolute_import
from __future__ import division
from past.utils import old_div
from .technique import PositionInvariantTechnique
import cv2
import numpy as np

class raiseGreenAugmentationTechnique(PositionInvariantTechnique):

    # Valid values for pover are in the range (0.25,4]
    def __init__(self,parameters):
        PositionInvariantTechnique.__init__(self, parameters)
        if 'power' in list(parameters.keys()):
            self.power = float(parameters["power"])
        else:
            self.power = 0.9

        if (self.power<=0.25 or self.power >4):
            raise NameError("Invalid value for power")

    def apply(self, image):
        if(len(image.shape)!=3):
            raise NameError("Not applicable technique")
        identityB = np.arange(256, dtype=np.dtype('uint8'))
        identityG = np.array([((old_div(i, 255.0)) ** self.power) * 255
                              for i in np.arange(0, 256)]).astype("uint8")
        identityR = np.arange(256, dtype=np.dtype('uint8'))
        lut = np.dstack((identityB, identityG, identityR))

        # apply gamma correction using the lookup table
        image = cv2.LUT(image, lut)
        return image


from __future__ import absolute_import
from builtins import str
from builtins import object
from .iaugmentor import IAugmentor
from imutils import paths
import os
import cv2
from sklearn.externals.joblib import Parallel, delayed

# We need to define this function outside to work in parallel.
def readAndGenerateImage(outputPath, generators, i_and_imagePath):
    (i, imagePath) = i_and_imagePath
    image = cv2.imread(imagePath)
    label = imagePath.split(os.path.sep)[-2]
    name = imagePath.split(os.path.sep)[-1]
    for (j, generator) in enumerate(generators):
        newimage = generator.applyForClassification(image)
        cv2.imwrite(outputPath + label + "/" + str(i) + "_" + str(j) + "_" + name,
                    newimage)

# This class serves to generate images for a classification
# problem where all the images are organized by folders
# distributed by labels. Example:
# - Folder
# |- cats
#    |- image1.jpg
#    |- image2.jpg
#    |- ...
# |- dogs
#    |- image1.jpg
#    |- image2.jpg
#    |- ...
class FolderLinearClassificationAugmentor(object):

    def __init__(self,inputPath,parameters):
        IAugmentor.__init__(self)
        self.inputPath = inputPath
        # output path represents the folder where the images will be stored
        if parameters["outputPath"]:
            self.outputPath = parameters["outputPath"]
        else:
            raise ValueError("You should provide an output path in the parameters")

        self.generators = []

    def addGenerator(self, generator):
        self.generators.append(generator)

    def readImagesAndAnnotations(self):
        self.imagePaths = list(paths.list_files(self.inputPath,validExts=(".jpg", ".jpeg", ".png", ".bmp",".tiff",".tif")))

    def createOutputDirs(self):
        dirs = [d for d in os.listdir(self.inputPath) if os.path.isdir(os.path.join(self.inputPath, d))]
        for dir in dirs:
            os.makedirs(self.outputPath + dir + "/")

    def applyAugmentation(self):
        self.readImagesAndAnnotations()
        self.createOutputDirs()
        len1 = len(self.imagePaths)
        Parallel(n_jobs=-1)(delayed(readAndGenerateImage)(self.outputPath,self.generators,x) for x in enumerate(self.imagePaths))


# # Example
# augmentor = FolderLinearClassificationAugmentor(
#     "/home/joheras/datasets/cats_and_dogs_small/train/",
#     "/home/joheras/datasets/cats_and_dogs_small/data-augmented-parallel/"
# )
#
# from techniques.averageBlurringAugmentationTechnique import averageBlurringAugmentationTechnique
# from techniques.bilateralBlurringAugmentationTechnique import bilateralBlurringAugmentationTechnique
# from techniques.gaussianNoiseAugmentationTechnique import gaussianNoiseAugmentationTechnique
# from techniques.rotateAugmentationTechnique import rotateAugmentationTechnique
# from techniques.flipAugmentationTechnique import flipAugmentationTechnique
# from techniques.noneAugmentationTechnique import noneAugmentationTechnique
# from generator import Generator
# import time
# augmentor.addGenerator(Generator(noneAugmentationTechnique()))
# augmentor.addGenerator(Generator(averageBlurringAugmentationTechnique()))
# augmentor.addGenerator(Generator(bilateralBlurringAugmentationTechnique()))
# augmentor.addGenerator(Generator(gaussianNoiseAugmentationTechnique()))
# augmentor.addGenerator(Generator(rotateAugmentationTechnique()))
# augmentor.addGenerator(Generator(flipAugmentationTechnique()))
# start = time.time()
# augmentor.applyAugmentation()
# end = time.time()
# print(end - start)



from __future__ import absolute_import
from .technique import PositionInvariantTechnique
import cv2
import numpy as np

class gaussianNoiseAugmentationTechnique(PositionInvariantTechnique):


    def __init__(self,parameters):
        PositionInvariantTechnique.__init__(self, parameters)
        if 'mean' in list(parameters.keys()):
            self.mean = float(parameters["mean"])
        else:
            self.mean = 0
        if 'sigma' in list(parameters.keys()):
            self.sigma = float(parameters["sigma"])
        else:
            self.sigma = 10

    def apply(self, image):
        im = np.zeros(image.shape, np.uint8)
        if(len(image.shape)==2):
            m = self.mean
            s = self.sigma
        else:
            m = (self.mean,self.mean,self.mean)
            s = (self.sigma,self.sigma,self.sigma)
        cv2.randn(im, m, s)
        image_noise = cv2.add(image, im)
        return image_noise






from __future__ import absolute_import
from __future__ import division
from builtins import str
from builtins import object
from past.utils import old_div
from .iaugmentor import IAugmentor
from imutils import paths
import os
import cv2
from sklearn.externals.joblib import Parallel, delayed
import xml.etree.ElementTree as ET
from .utils import prettify



def readAndGenerateImage(outputPath, generators, i_and_imagePath):

    (i, imagePath) = i_and_imagePath
    image = cv2.imread(imagePath)
    (hI, wI) = image.shape[:2]
    name = imagePath.split(os.path.sep)[-1]
    labelPath = '/'.join(imagePath.split(os.path.sep)[:-1]) + "/"+name[0:name.rfind(".")] + ".txt"
    lines = [line.rstrip('\n') for line in open(labelPath)]
    #if(len(objects)<1):
    #    raise Exception("The xml should contain at least one object")
    boxes = []
    if lines != ['']:
        for line in lines:
            components = line.split(" ")
            category = components[0]
            x  = int(float(components[1])*wI - float(components[3])*wI/2)
            y = int(float(components[2])*hI - float(components[4])*hI/2)
            h = int(float(components[4])*hI)
            w = int(float(components[3])*wI)
            boxes.append((category, (x, y, w, h)))
    for (j, generator) in enumerate(generators):
        (newimage, newboxes) = generator.applyForDetection(image, boxes)
        (hI, wI) = newimage.shape[:2]

        if newboxes is not None:
            cv2.imwrite(outputPath + "/" + str(i) + "_" + str(j) + "_" + name,
                        newimage)


            file = open(outputPath + "/" + str(i) + "_" + str(j) + "_" + name[0:name.rfind(".")]+".txt", "w")
            for box in newboxes:
                (category, (x, y, wb, hb)) = box
                file.write(category + " " + str(old_div(float(x+old_div(wb,2)),wI)) + " " + str(old_div(float(y+old_div(hb,2)),hI)) + " " + str(old_div(float(wb),wI)) + " " + str(old_div(float(hb),hI))+ "\n")
            file.close()
        else:
            file = open(outputPath + "/" + str(i) + "_" + str(j) + "_" + name[0:name.rfind(".")] + ".txt", "w")
            file.close()


    #
# # This class serves to generate images for a detection
# # problem where all the images in the given folder using the JPG format, and the labels
# # are given in the same folder with the same name and using the YOLO format.
# # Example:
# # - Folder
# # |- image1.jpg
# # |- image1.txt
# # |- image2.jpg
# # |- image2.txt
# # |- ...
# #
#
class yoloLinearDetectionAugmentor(object):

    def __init__(self,inputPath,parameters):
        IAugmentor.__init__(self)
        self.inputPath = inputPath
        # output path represents the folder where the images will be stored
        if parameters["outputPath"]:
            self.outputPath = parameters["outputPath"]
        else:
            raise ValueError("You should provide an output path in the parameters")

        self.generators = []

    def addGenerator(self, generator):
        self.generators.append(generator)

    def readImagesAndAnnotations(self):
        self.imagePaths = list(paths.list_files(self.inputPath,validExts=(".jpg", ".jpeg")))
        self.labelPaths = list(paths.list_files(self.inputPath,validExts=(".txt")))
        if (len(self.imagePaths) != len(self.labelPaths)):
            raise Exception("The number of images is different to the number of annotations")

    def applyAugmentation(self):
        self.readImagesAndAnnotations()
        Parallel(n_jobs=-1)(delayed(readAndGenerateImage)(self.outputPath,self.generators,x) for x in enumerate(self.imagePaths))


#
# # Example
# augmentor = yoloLinearDetectionAugmentor(
#     "/home/joheras/datasets/estomas-yolo/",
#     {"outputPath": "/home/joheras/datasets/estomas-yolo/"}
# )
#
# from ..techniques.averageBlurringAugmentationTechnique import averageBlurringAugmentationTechnique
# from techniques.bilateralBlurringAugmentationTechnique import bilateralBlurringAugmentationTechnique
# from techniques.gaussianNoiseAugmentationTechnique import gaussianNoiseAugmentationTechnique
# from techniques.rotateAugmentationTechnique import rotateAugmentationTechnique
# from ..techniques.flipAugmentationTechnique import flipAugmentationTechnique
# from techniques.noneAugmentationTechnique import noneAugmentationTechnique
# from generator import Generator
# import time
# augmentor.addGenerator(Generator(noneAugmentationTechnique()))
# augmentor.addGenerator(Generator(averageBlurringAugmentationTechnique()))
# augmentor.addGenerator(Generator(bilateralBlurringAugmentationTechnique()))
# augmentor.addGenerator(Generator(gaussianNoiseAugmentationTechnique()))
# augmentor.addGenerator(Generator(rotateAugmentationTechnique()))
# augmentor.addGenerator(Generator(flipAugmentationTechnique()))
# start = time.time()
# augmentor.applyAugmentation()
# end = time.time()
# print(end - start)



from __future__ import absolute_import
from .techniqueList import  Techniques

def createTechnique(technique,parameters):
    Technique = Techniques[technique]
    if Technique is None:
        raise ValueError("That technique is not available")
    return Technique(parameters)


from __future__ import absolute_import
from .technique import PositionInvariantTechnique
import cv2
import numpy as np

class changeToLABAugmentationTechnique(PositionInvariantTechnique):


    def __init__(self,parameters):
        PositionInvariantTechnique.__init__(self, parameters, False)


    def apply(self, image):

        image1 = cv2.cvtColor(image,cv2.COLOR_BGR2LAB)
        return image1


from __future__ import absolute_import
from .averageBlurringAugmentationTechnique import averageBlurringAugmentationTechnique
from .bilateralBlurringAugmentationTechnique import bilateralBlurringAugmentationTechnique
from .blurringAugmentationTechnique import blurringAugmentationTechnique
from .cropAugmentationTechnique import cropAugmentationTechnique
from .dropoutAugmentationTechnique import dropoutAugmentationTechnique
from .elasticTransformAugmentationTechnique import elasticTransformAugmentationTechnique
from .equalizeHistogramAugmentationTechnique import equalizeHistogramAugmentationTechnique
from .flipAugmentationTechnique import flipAugmentationTechnique
from .gammaCorrectionAugmentationTechnique import gammaCorrectionAugmentationTechnique
from .gaussianBlurringAugmentationTechnique import gausianBlurringAugmentationTechnique
from .gaussianNoiseAugmentationTechnique import gaussianNoiseAugmentationTechnique
from .medianBlurringAugmentationTechnique import medianBlurringAugmentationTechnique
from .noneAugmentationTechnique import noneAugmentationTechnique
from .raiseBlueAugmentationTechnique import raiseBlueAugmentationTechnique
from .raiseGreenAugmentationTechnique import raiseGreenAugmentationTechnique
from .raiseHueAugmentationTechnique import raiseHueAugmentationTechnique
from .raiseRedAugmentationTechnique import raiseRedAugmentationTechnique
from .raiseSaturationAugmentationTechnique import raiseSaturationAugmentationTechnique
from .raiseValueAugmentationTechnique import raiseValueAugmentationTechnique
from .resizeAugmentationTechnique import resizeAugmentationTechnique
from .rotateAugmentationTechnique import rotateAugmentationTechnique
from .saltAndPepperNoiseAugmentationTechnique import saltAndPepperNoiseAugmentationTechnique
from .sharpenAugmentationTechnique import sharpenAugmentationTechnique
from .shearingAugmentationTechnique import shearingAugmentationTechnique
from .translationAugmentationTechnique import translationAugmentationTechnique
from .shiftChannelAugmentationTechnique import shiftChannelAugmentationTechnique
from .changeToHSVAugmentationTechnique import changeToHSVAugmentationTechnique
from .changeToLABAugmentationTechnique import changeToLABAugmentationTechnique
from .invertAugmentationTechnique import invertAugmentationTechnique
from .cropSizeAugmentationTechnique import cropSizeAugmentationTechnique

Techniques = {
    "average_blurring" : averageBlurringAugmentationTechnique,
    "bilateral_blurring" : bilateralBlurringAugmentationTechnique,
    "blurring" : blurringAugmentationTechnique,
    "change_to_hsv" : changeToHSVAugmentationTechnique,
    "change_to_lab" : changeToLABAugmentationTechnique,
    "crop" : cropAugmentationTechnique,
    "cropSize" : cropSizeAugmentationTechnique,
    "dropout" : dropoutAugmentationTechnique,
    "elastic" : elasticTransformAugmentationTechnique,
    "equalize_histogram" : equalizeHistogramAugmentationTechnique,
    "flip" : flipAugmentationTechnique,
    "gamma" : gammaCorrectionAugmentationTechnique,
    "gaussian_blur": gausianBlurringAugmentationTechnique,
    "gaussian_noise": gaussianNoiseAugmentationTechnique,
    "invert": invertAugmentationTechnique,
    "median_blur": medianBlurringAugmentationTechnique,
    "none":  noneAugmentationTechnique,
    "raise_blue":  raiseBlueAugmentationTechnique,
    "raise_green":raiseGreenAugmentationTechnique,
    "raise_hue": raiseHueAugmentationTechnique,
    "raise_red": raiseRedAugmentationTechnique,
    "raise_saturation": raiseSaturationAugmentationTechnique,
    "raise_value":raiseValueAugmentationTechnique,
    "resize": resizeAugmentationTechnique,
    "rotate": rotateAugmentationTechnique,
    "salt_and_pepper":saltAndPepperNoiseAugmentationTechnique,
    "sharpen":sharpenAugmentationTechnique,
    "shift_channel":shiftChannelAugmentationTechnique,
    "shearing": shearingAugmentationTechnique,
    "translation": translationAugmentationTechnique
}



from builtins import object
from abc import ABCMeta, abstractmethod
from future.utils import with_metaclass

class Technique(with_metaclass(ABCMeta, object)):
    def __init__(self, parameters = None):
        self.parameters = parameters

    @abstractmethod
    def apply(self, image):
        raise NotImplementedError


class PositionVariantTechnique(with_metaclass(ABCMeta, Technique)):
    def __init__(self, parameters=None):
        Technique.__init__(self, parameters)

    @abstractmethod
    def apply(self, image):
        raise NotImplementedError

class PositionInvariantTechnique(with_metaclass(ABCMeta, Technique)):
    def __init__(self, parameters=None):
        Technique.__init__(self, parameters)

    @abstractmethod
    def apply(self, image):
        raise NotImplementedError


from builtins import object
from abc import ABCMeta, abstractmethod
from future.utils import with_metaclass

class ITechnique(with_metaclass(ABCMeta, object)):
    def __init__(self, parameters = None,changeLabel=False):
        self.changeLabel = changeLabel
        self.parameters = parameters

    @abstractmethod
    def apply(self, image):
        raise NotImplementedError


from __future__ import absolute_import
from builtins import object
from .detection import detectBox,detectBoxes
from ..techniques.technique import PositionVariantTechnique,PositionInvariantTechnique


class Generator(object):

    def __init__(self,technique):
        if (isinstance(technique, PositionVariantTechnique)) or (isinstance(technique, PositionInvariantTechnique)):
            self.technique = technique
        else:
            raise ValueError("The technique should be an object of the class Technique")

    def applyForClassification(self, image):
        return self.technique.apply(image)


    def applyForLocalization(self,image,box):
        if(isinstance(self.technique,PositionVariantTechnique)):
            return [self.technique.apply(image),
                    detectBox(image,box,self.technique.apply)]
        else:
            return [self.technique.apply(image),box]

    def applyForDetection(self, image, boxes):
        if (isinstance(self.technique, PositionVariantTechnique)):
            return [self.technique.apply(image),
                    detectBoxes(image, boxes, self.technique.apply)]
        else:
            return [self.technique.apply(image), boxes]

    def applyForSegmentation(self, image, imageLabel):
        if (isinstance(self.technique, PositionVariantTechnique)):
            return [self.technique.apply(image),
                    self.technique.apply(imageLabel)]
        else:
            return [self.technique.apply(image), imageLabel]





import cv2
import numpy as np
import imutils

# A box is a tuple (category,(x,y,w,h)) where category is the category of the
# object, x and y represent the top-left corner, w is the width and h is the height.
def detectBox(image,box,function):
    mask = np.zeros(image.shape[:2], dtype="uint8")
    (category,(x,y,w,h)) = box
    cv2.rectangle(mask, (x, y), (x+w, y+h), 255, -1)
    newmask = function(*[mask])

    cnts = cv2.findContours(newmask.copy(), cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
    cnts = cnts[0] if imutils.is_cv2() else cnts[1]
    if(len(cnts)==0):
        return None
    return (category,cv2.boundingRect(cnts[0]))


# Boxes is a list of boxes with the following format: (category,(x,y,w,h))
def detectBoxes(image,boxes,function):
    return [detectBox(image,box,function) for box in boxes if detectBox(image,box,function) is not None]








